@@ -1,11 +1,14 @@
+import os
 import sys
 import json
 import traceback
 import importlib.util
 import time
 import threading
+import queue
 import datetime
 import multiprocessing
+import pg_tasks_queue.Kthread as kthread
 from pg_tasks_queue.Database import TasksDatabase as database
 
 
@@ -16,7 +19,7 @@ class Worker(object):
     _started = False
     _sleep_sec = 1.
 
-    def init(self, config_dict, keep_connection=False):
+    def init(self, config_dict, connection=None, schema=None, keep_connection=False):
         func_name = f'{self.__class__.__name__}.{sys._getframe().f_code.co_name}()'
         try:
             if not isinstance(config_dict, dict):
@@ -26,7 +29,10 @@ class Worker(object):
             if not isinstance(database_dict, dict):
                 print(f'Error in {func_name}: not isinstance(database_dict, dict)...')
                 return False
-            kwargs = {'settings': database_dict, 'auto_disconnect': False if keep_connection else True}
+            kwargs = {'settings': database_dict,
+                      'connection': connection,
+                      'schema': schema,
+                      'auto_disconnect': False if (connection or keep_connection) else True}
             if not database.init(**kwargs):
                 print(f'Error in {func_name}: not database.init(); return...')
                 return False
@@ -78,89 +84,133 @@ class Worker(object):
             print(f'Error in {func_name}: {type(e)}: {str(e)}; traceback: {traceback.print_exc()}')
             return None
 
-    def worker_process(self, queue, task_dict):
+    def worker(self, task_dict):
         func_name = f'{self.__class__.__name__}.{sys._getframe().f_code.co_name}()'
-        # print(f'{func_name}; os.getpid(): {os.getpid()}; os.getppid(): {os.getppid()}; task_dict: {task_dict}')
-        task_module = task_dict.get('module')
-        task_func = task_dict.get('func')
-        func = self.import_function_from_module(task_module, task_func)
-        if func is None:
-            error = f'Error in {func_name}: import_function_from_module({task_module}, {task_func}) is None'
-            print(error)
-            result = {'error': error}
-        else:
-            params = task_dict.get('params')
-            if isinstance(params, str):
-                try:
+        print(f'{func_name}; os.getpid(): {os.getpid()}; os.getppid(): {os.getppid()}')
+        try:
+            task_module = task_dict.get('module')
+            task_func = task_dict.get('func')
+            func = self.import_function_from_module(task_module, task_func)
+            if func is None:
+                error = f'Error in {func_name}: import_function_from_module({task_module}, {task_func}) is None'
+                print(error)
+                return {'error': error}
+            else:
+                params = task_dict.get('params')
+                if isinstance(params, str):
                     params = json.loads(params)
                     if not isinstance(params, dict):
                         params = dict()
-                    # print(f'{func_name}: try to start "{task_module}.{task_func}"...')
-                    result = func(**params)
-                except Exception as e:
-                    error = f'Error in {func_name}: {type(e)}: {str(e)}; traceback: {traceback.print_exc()}'
-                    print(error)
-                    result = {'status': 'error', 'result': error}
+                    return func(**params)
+        except Exception as e:
+            error = f'Error in {func_name}: {type(e)}: {str(e)}; traceback: {traceback.print_exc()}'
+            print(error)
+            return {'status': 'error', 'result': error}
+
+    def worker_process(self, queue, task_dict):
+        result = self.worker(task_dict)
         queue.put(result)
 
-    def _do_task(self):
+    def update_db_task(self, task_dict, result_dict):
         func_name = f'{self.__class__.__name__}.{sys._getframe().f_code.co_name}()'
-        # print(f'{func_name}; os.getpid(): {os.getpid()}; os.getppid(): {os.getppid()}')
+        if not isinstance(result_dict, dict):
+            error = f'Error in {func_name}: not isinstance(_result_dict, dict)...'
+            print(error)
+            database.update_task_error(task_dict, error)
+        else:
+            status = result_dict.get('status')
+            result = result_dict.get('result', '')
+            if not isinstance(status, str):
+                error = f'Error in {func_name}: not isinstance(status, str)...'
+                print(error)
+                database.update_task_error(task_dict, error)
+            elif status == 'error':
+                database.update_task_error(task_dict, result)
+            else:
+                update_values = {'finished_time': datetime.datetime.now(),
+                                 'status': status,
+                                 'result': result}
+                database.update_task(task_dict.get('task_id'), update_values)
+
+    def stop_woker(self, func_name, task_dict):
+        error = f'Error in {func_name}: worker.is_alive() => os.kill(os.getpid(), signal.SIGINT)'
+        database.update_task_error(task_dict, error)
+        try:
+            # sys.exit(1)
+            import signal
+            os.kill(os.getpid(), signal.SIGINT)
+        finally:
+            print(error)
+
+    def _do_task(self, process_type):
+        func_name = f'{self.__class__.__name__}.{sys._getframe().f_code.co_name}()'
+        print(f'{func_name}; os.getpid(): {os.getpid()}; os.getppid(): {os.getppid()}')
         if not self._started:
             print(f'{func_name}; self._started = False; return')
             return
-
         try:
             task_dict = database.get_one_task()
             # print(f'{func_name}: now={datetime.datetime.now()}; task_dict: {task_dict}')
             if isinstance(task_dict, dict):
-                queue = multiprocessing.Queue()
-                worker = multiprocessing.Process(target=self.worker_process, args=(queue, task_dict,))
-                worker.start()
-                worker.join(timeout=self._timeout_sec)
-                if worker.is_alive():
-                    worker.terminate()
-                    database.update_task_error(task_dict, f'Error in {func_name}: worker.terminate()')
-                else:
-                    if not queue.empty():
-                        queue_get = queue.get_nowait()
-                        if not isinstance(queue_get, dict):
-                            error = f'Error in {func_name}: not isinstance(queue_get, dict)...'
+                if process_type is None or process_type.lower() == 'none' or process_type.lower() == '<none>':
+                    timer = threading.Timer(self._timeout_sec, self.stop_woker, args=(func_name, task_dict,))
+                    timer.setDaemon(True)
+                    timer.start()
+                    result_dict = self.worker(task_dict)
+                    timer.cancel()
+                    self.update_db_task(task_dict, result_dict)
+                elif process_type in ['fork', 'thread', 'kthread']:
+                    if process_type == 'fork':
+                        worker_queue = multiprocessing.Queue()
+                        worker = multiprocessing.Process(target=self.worker_process, args=(worker_queue, task_dict,))
+                    else:
+                        worker_queue = queue.Queue()
+                        if process_type == 'thread':
+                            worker = threading.Thread(target=self.worker_process, args=(worker_queue, task_dict,))
+                            worker.setDaemon(True)
+                        else:
+                            worker = kthread.KThread(target=self.worker_process, args=(worker_queue, task_dict,))
+                    worker.start()
+                    worker.join(timeout=self._timeout_sec)
+                    if worker.is_alive():
+                        if process_type in ['fork', 'kthread']:
+                            worker.terminate()
+                            error = f'Error in {func_name}: worker({process_type}).is_alive() => worker.terminate()'
                             print(error)
                             database.update_task_error(task_dict, error)
                         else:
-                            status = queue_get.get('status')
-                            result = queue_get.get('result', '')
-                            if not isinstance(status, str):
-                                error = f'Error in {func_name}: not isinstance(queue_get, dict)...'
-                                print(error)
-                                database.update_task_error(task_dict, error)
-                            elif status == 'error':
-                                database.update_task_error(task_dict, result)
-                            else:
-                                update_values = {'finished_time': datetime.datetime.now(),
-                                                 'status': status,
-                                                 'result': result}
-                                database.update_task(task_dict.get('task_id'), update_values)
+                            error = f'Error in {func_name}: worker({process_type}).is_alive() => sys.exit(1)'
+                            print(error)
+                            database.update_task_error(task_dict, error)
+                            sys.exit(1)
                     else:
-                        error = f'Error in {func_name}: queue is empty()...'
-                        print(error)
-                        database.update_task_error(task_dict, error)
+                        if not worker_queue.empty():
+                            self.update_db_task(task_dict, worker_queue.get_nowait())
+                        else:
+                            error = f'Error in {func_name}: queue is empty()...'
+                            print(error)
+                            database.update_task_error(task_dict, error)
 
             if not self._blocking:
                 threading.Timer(self._sleep_sec, self._do_task).start()
-
         except Exception as e:
             print(f'Error in {func_name}: {type(e)}: {str(e)}; traceback: {traceback.print_exc()}')
 
-    def start(self, blocking=True, sleep_sec=None):
+    def start(self, blocking=True, sleep_sec=None, process_type=None, timeout_sec=None):
+        func_name = f'{self.__class__.__name__}.{sys._getframe().f_code.co_name}()'
+        if process_type is not None:
+            if process_type.lower() not in ['fork', 'thread', 'kthread', 'none', '<none>']:
+                print(f'Error in {func_name}: unknown process_type "{process_type}"...')
+                return
         self._blocking = blocking
         if isinstance(sleep_sec, float):
             self._sleep_sec = sleep_sec
+        if isinstance(timeout_sec, float):
+            self._timeout_sec = timeout_sec
         self._started = True
         if self._blocking:
             while self._started:
-                self._do_task()
+                self._do_task(process_type)
                 time.sleep(self._sleep_sec)
         else:
             self._do_task()
